##################################################################################
#
#   Sketchup to Marathon Export Script - Terminals
#
##################################################################################

module Marathon

	DebugTerminal = false

	#---------------------------------------------------------------------------------

	class TerminalError < StandardError
	end

	#---------------------------------------------------------------------------------

	class MarathonTerminalGroup
	
		attr :flags, true
		attr :type, true
		attr :permutation, true
		attr :start_index, true
		attr :length, true
		attr :maximum_line_count, true
		
		def initialize()
			@maximum_line_count = 1
		end
	
	end # class MarathonTerminalGroup
	
	#---------------------------------------------------------------------------------

	class MarathonTerminalFont
	
		attr :index, true
		attr :face, true
		attr :color, true
	
	end # class MarathonTerminalFont

	#---------------------------------------------------------------------------------

	class MarathonTerminal
	
		LINES_PER_PAGE = 22 # TODO: Make this configurable?
	
		attr :number
		attr :filename
		attr :representative_line
		attr :flags
		attr :groups
		attr :font_changes
		
		def initialize(n, name, line)
			@number = n
			@filename = name
			@representative_line = line
			@flags = 0
			@groups = []
			@font_changes = []
			@text_chunks = []
			@text_length = 0
		end
		
		def total_length()
			length = 10 + 12 * @groups.length + 6 * @font_changes.length
			for text in @text_chunks
				length += text.length
			end
			return length
		end
		
		def group_count
			return @groups.length
		end
		
		def font_changes_count
			return @font_changes.length
		end
		
		def lines_per_page
			return LINES_PER_PAGE
		end
		
		def text
			return @text_chunks.join()
		end
		
		def parse_file()
			if DebugTerminal
				printf("Scanning terminal file %s\n", @filename.inspect)
			end
			dir = File.dirname(Sketchup.active_model.path)
			path = File.join(dir, @filename)
			scanner = TerminalScanner.new(path)
			parser = TerminalParser.new(self)
			parser.parse(scanner)
			self.finish_last_group()
		end
		
		def add_group(type, permutation)
			self.finish_last_group()
			group = MarathonTerminalGroup.new()
			group.flags = 0 # TODO: Alignment options
			group.type = type
			group.permutation = permutation
			group.start_index = @text_length
			@groups.push(group)
		end
		
		def add_font_change(color, style)
			font = MarathonTerminalFont.new()
			font.index = @text_length
			font.face = style
			font.color = color
			@font_changes.push(font)
		end
		
		def add_text(text)
			@text_chunks.push(text.gsub("\n", "\r"))
			@text_length += text.length
			group = @groups[-1]
			if group
				n = text.count("\n")
				group.maximum_line_count += n
			end
		end
		
		def finish_last_group()
			if @groups.length > 0
				self.add_text("\000")
				group = @groups[-1]
				group.length = @text_length - group.start_index
			end
		end
		
		def dump()
			printf("MarathonTerminal %s:\n", @number)
			for i in 0...@groups.length
				printf("G%s: %s\n", i, @groups[i].inspect)
			end
			for i in 0...@font_changes.length
				printf("F%s: %s\n", i, @font_changes[i].inspect)
			end
			i = 0
			for chunk in @text_chunks
				printf("T%s: %s\n", i, chunk.inspect)
				i += chunk.length
			end
		end
	
	end # class MarathonTerminal

	#---------------------------------------------------------------------------------

	TagTable = {
		"logon" => [:Group, 0],
		"unfinished" => [:Group, 1],
		"success" => [:Group, 2],
		"failure" => [:Group, 3],
		"information" => [:Group, 4],
		"end" => [:Group, 5],
		"teleport" => [:Group, {
			"level" => 6, # permutation is level to go to
			"polygon" => 7}], # permutation is polygon to go to.
		"checkpoint" => [:Group, 8], # permutation is the goal to show
		"sound" => [:Group, 9], # permutation is the sound id to play
		"movie" => [:Group, 10], # permutation is the movie id to play
		"track" => [:Group, 11], # permutation is the track to play
		"pict" => [:Group, 12], # permutation is the pict to display
		"logoff" => [:Group, 13],
		"camera" => [:Group, 14], #  permutation is the object index
		"static" => [:Group, 15], # permutation is the duration of static.
		"tag" => [:Group, 16], # permutation is the tag to activate
		"ltgreen" => [:Color, 0],
		"white" => [:Color, 1],
		"red" => [:Color, 2],
		"dkgreen" => [:Color, 3],
		"ltblue" => [:Color, 4],
		"yellow" => [:Color, 5],
		"dkred" => [:Color, 6],
		"dkblue" => [:Color, 7],
		"b" => [:Style, 1],
		"i" => [:Style, 2],
		"u" => [:Style, 4],
		"/b" => [:Style, -1],
		"/i" => [:Style, -2],
		"/u" => [:Style, -4],
	}
	
	#---------------------------------------------------------------------------------

	class TerminalParser
	
		def initialize(term)
			@term = term
			@current_color = 0
			@current_style = 0
			@beginning_of_group = false
		end
		
		def parse(scanner)
			@scanner = scanner
			for item in scanner
				if DebugTerminal
					printf("Parser: Processing item %s\n", item.inspect) ###
				end
				if item[0, 1] == "<" and item.length > 1
					self.process_tag(item)
				else
					self.add_text(item)
				end
			end
		end
		
		def process_tag(tag)
			tag_name, param_name, param_value = self.parse_tag(tag)
			entry = TagTable[tag_name]
			if entry
				kind, code = entry
				if code.instance_of?(Hash)
					code = code[param_name]
					if not code
						if param_name
							@scanner.error(format("Unknown parameter '%s'", param_name))
						else
							@scanner.error(format("Parameter required for directive '%s'", tag_name))
						end
					end
				end
				case kind
					when :Group
						self.add_group(code, param_value)
						@beginning_of_group = true
					when :Color
						self.change_color(code)
					when :Style
						self.change_style(code)
				end
			else
				@scanner.error(format("Unknown directive '%s'", tag_name))
			end
		end
		
		TagPattern = /<\s*(\/?\w+)(\s+((\w+)\s*=\s*)?"?(\w+)"?)?\s*>/

		def parse_tag(tag)
			match = TagPattern.match(tag)
			if not match
				@scanner.error(format("Syntax error in tag: %s", tag))
			end
			result = [match[1], match[4], match[5]]
			if DebugTerminal
				printf("Tag %s parsed into %s\n", tag, result.inspect)
			end
			return result
		end
		
		def add_group(group_type, param_value)
			@term.add_group(group_type, param_value)
		end
		
		def change_color(color)
			@current_color = color
			@term.add_font_change(@current_color, @current_style)
		end
		
		def change_style(bit)
			if bit > 0
				@current_style |= bit
			else
				@current_style &= ~(-bit)
			end
			@term.add_font_change(@current_color, @current_style)
		end
		
		def add_text(item)
			if @beginning_of_group
				if item[-1, 1] == "\n"
					@beginning_of_group = false
				end
				if item =~ /\s+/
					if DebugTerminal
						printf("Ignoring whitespace after group tag\n")
					end
					return
				end
			end
			@term.add_text(item)
		end	

	end # class TerminalParser
		
	#---------------------------------------------------------------------------------

	class TerminalScanner
	
		Tokens = Regexp.new([
			"\\\\<",
			"\\\\\\\\",
			"</?[^>]*>",
		].join("|"))
	
		def initialize(filename)
			@filename = filename
			@lineno = 0
			if DebugTerminal
				printf("Opening terminal file %s\n", filename.inspect)
			end
			@f = File.open(filename)
			#printf("File = %s\n", @f.inspect) ###
		end
		
		def each()
			while line = @f.gets()
				@lineno += 1
				while match = Tokens.match(line)
					pre = match.pre_match
					if pre != ""
						yield pre
					end
					tok = match.to_s
					case tok
						when "\\<"
							tok = "<"
						when "\\\\"
							tok = "\\"
					end
					yield tok
					line[0, match.end(0)] = ""
				end
				if line != ""
					yield line
				end
			end
		end
		
		def error(mess)
			raise TerminalError.new(format("%s, line %s: %s",
				@filename, @lineno, mess))
		end
	
	end # class TerminalScanner

	#---------------------------------------------------------------------------------

	### Testing
	#s = MarathonTerminal.new(1, "test.term", nil)
	#s.parse_file()
	#s.dump()

end # module Marathon
