###############################################################################
#
#   MarathUp - Property access and editing utilities
#
###############################################################################

module Marathon

	TransferModes = [
		 "Normal",
		 "Fade to black",
		 "Invisible",
		 "Sub-invisible",
		 "Pulsate (poly only)",
		 "Wobble (poly only)",
		 "Fast wobble (poly only)",
		 "Static",
		 "Sub-static",
		 "Landscape",
		 "Smear (flat shading)",
		 "Fade out static",
		 "Pulsate static",
		 "Fold in",
		 "Fold out",
		 "Horizontal slide",
		 "Fast horizontal slide",
		 "Vertical slide",
		 "Fast vertical slide",
		 "Wander",
		 "Fast wander"
	]

	def Marathon.abort(fmt, *args)
		UI.beep
		UI.messagebox(format(fmt, *args))
		throw :Abort
	end

	#----------------------------------------------------------------------------------
	#
	#   Testing the type of entities
	#
	#----------------------------------------------------------------------------------

	def Marathon.classify_face(face)
		if face.instance_of?(Sketchup::Face)
			normal = face.normal
			z = normal.z
			if z.abs < 0.01
				return :Wall
			elsif z < -0.99
				return :Floor
			elsif z > 0.99
				return :Ceiling
			end
		end
		return :Unknown
	end
	
	def Marathon.is_floor_face(face)
		return classify_face(face) == :Floor
	end

	def Marathon.is_ceiling_face(face)
		return classify_face(face) == :Ceiling
	end
	
	def Marathon.is_floor_or_ceiling_face(face)
		kind = classify_face(face)
		case kind
			when :Floor, :Ceiling
				return true
			else
				return false
		end
	end

	def Marathon.is_wall_face(face)
		return classify_face(face) == :Wall
	end

	def Marathon.object_kind(obj)
		desc = obj.definition.description
		if desc[0, 1] == "*"
			return desc.split(",")[0]
		else
			return nil
		end
	end
	
	def Marathon.is_group_or_instance(obj)
		return (obj.is_a?(Sketchup::ComponentInstance) \
			or obj.is_a?(Sketchup::Group))
	end
	
	def Marathon.is_light(obj)
		if not obj.instance_of?(Sketchup::ComponentInstance)
			return false
		end
		return obj.definition.description == "*Light"
	end

	def Marathon.is_media(obj)
		if not obj.instance_of?(Sketchup::ComponentInstance)
			return false
		end
		return obj.definition.description == "*Liquid"
	end
	
	#----------------------------------------------------------------------------------
	#
	#   Getting specific kinds of selections from the model
	#
	#----------------------------------------------------------------------------------

	def Marathon.get_single_selection()
		sel = Sketchup.active_model.selection
		if not sel.single_object?()
			abort("A single entity must be selected.")
		end
		return sel.first
	end
	
	def Marathon.get_selected_entities(desc, &filter)
		result = []
		for entity in Sketchup.active_model.selection
			if filter.call(entity)
				result.push(entity)
			end
		end
		if result.length == 0
			abort("No %s selected.", desc)
		end
		return result
	end

	def Marathon.get_selected_floor_face()
		obj = get_single_selection()
		if not is_floor_face(obj)
			abort("A floor face must be selected.")
		end
		return obj
	end
	
	def Marathon.get_selected_floor_faces()
		return get_selected_entities("floor faces") {|x| is_floor_face(x)}
	end
	
	def Marathon.get_selected_ceiling_face()
		obj = get_single_selection()
		if not is_ceiling_face(obj)
			abort("A ceiling face must be selected.")
		end
		return obj
	end
	
	def Marathon.get_selected_ceiling_faces()
		return get_selected_entities("ceiling faces") {|x| is_ceiling_face(x)}
	end
	
	def Marathon.get_selected_floor_or_ceiling_face()
		obj = get_single_selection()
		if not is_floor_or_ceiling_face(obj)
			abort("A ceiling face must be selected.")
		end
		return obj
	end
	
	def Marathon.get_selected_floor_or_ceiling_faces()
		return get_selected_entities("floor or ceiling faces") {
			|x| is_floor_or_ceiling_face(x)}
	end
	
	def Marathon.get_selected_wall_face()
		obj = get_single_selection()
		if not is_wall_face(obj)
			abort("A wall face must be selected.")
		end
		return obj
	end
	
	def Marathon.get_selected_wall_faces()
		return get_selected_entities("wall faces") {|x| is_wall_face(x)}
	end
	
	def Marathon.get_selected_instance()
		obj = get_single_selection()
		if not obj.instance_of?(Sketchup::ComponentInstance)
			abort("A component instance must be selected.")
		end
		return obj
	end
	
	def Marathon.get_selected_group_or_instance()
		obj = get_single_selection()
		if not (obj.is_a?(Sketchup::ComponentInstance) or obj.is_a?(Sketchup::Group))
			abort("A group or component instance must be selected.")
		end
		return obj
	end

#	def Marathon.get_selected_object()
#		obj = get_single_selection()
#		if not is_object(obj)
#			abort("A placeable object must be selected.")
#		end
#		return obj
#	end

	def Marathon.get_selected_groups_and_instances()
		kind = :Unknown
		result = get_selected_entities("objects") {|x| is_group_or_instance(x)}
		for obj in result
			if obj.is_a?(Sketchup::Group)
				k = nil
			else
				k = object_kind(obj)
			end
			if kind == :Unknown
				 kind = k
			elsif kind != k
				abort("More than one kind of object selected.")
			end
		end
		return kind, result
	end
	
	def Marathon.get_selected_light()
		obj = get_single_selection()
		if not is_light(obj)
			abort("A light must be selected.")
		end
		return obj
	end
	
	def Marathon.get_selected_media()
		obj = get_single_selection()
		if not is_media(obj)
			abort("A liquid must be selected.")
		end
		return obj
	end
	
	#----------------------------------------------------------------------------------
	#
	#   Getting and updating attribute dictionaries
	#
	#----------------------------------------------------------------------------------

	#  Attribute dictionary names
	
	FloorCeilingAttrs = "floor_ceiling"
	PolygonAttrs = "polygon"
	PlatformAttrs = "platform"
	SideAttrs = "side"
	ObjectAttrs = "object"
	LightAttrs = "light"
	MediaAttrs = "media"
	AmbientSoundAttrs = "ambient_sound"
	RandomSoundAttrs = "random_sound"
	NoteAttrs = "note"
	LevelAttrs = "map_info"
	MonsterPlacementAttrs = "monster_placement"
	ItemPlacementAttrs = "item_placement"
	ParameterAttrs = "parameters"
	
	def Marathon.get_attributes(entity, key, create)
		#  Get an attribute dictionary. If the dictionary does not already
		#  exist, it is created if create is true, otherwise an empty
		#  dictionary is returned. 
		attrs = entity.attribute_dictionary(key, create)
		if not attrs
			attrs = {}
		end
		return attrs
	end
	
	def Marathon.get_multiple_attributes(entities, key)
		#  Construct a merged attribute dictionary from the attribute
		#  dictionaries of a set of entities, containing :Mixed
		#  where the values are not the same across all entities.
		#  Flags attributes are represented as [value, mask] where
		#  mask has 1s in the place of mixed values.
		merged = {}
		for entity in entities
			attrs = entity.attribute_dictionary(key)
			if attrs
				for name, value in attrs
					#printf("Merging value for %s last 5 %s\n", name, name[-5, 5]) ###
					if name[-5, 5] == "flags"
						merged[name] = merge_flags(merged[name], value)
					elsif not merged.has_key?(name)
						merged[name] = value
					elsif merged[name] != value
						merged[name] = :Mixed
					end
				end
			end
		end
		return merged
	end
	
	def Marathon.merge_flags(existing, new)
		#printf("Merging flags %s with %s", existing.inspect, new.inspect) ###
		if not existing
			result = [new, 0]
		else
			existing[1] |= (existing[0] ^ new)
			result = existing
		end
		#printf(" --> %s\n", result.inspect) ###
		return result
	end
	
	def Marathon.set_multiple_attributes(entities, key, merged)
		#  Update the attributes dictionaries of a set of entities
		#  with the supplied values, where the value is not :Mixed.
		values = {}
		for name, value in merged
			if value != :Mixed
				values[name] = value
			end
		end
		if values.length > 0
			for entity in entities
				attrs = entity.attribute_dictionary(key, true)
				for name, value in values
					if name[-5, 5] == "flags"
						attrs[name] = update_flags(attrs[name], value)
					elsif value
						attrs[name] = value
					else
						attrs.delete_key(name)
					end
				end
			end
		end
	end
	
	def Marathon.update_flags(existing, new)
		#printf("Updating flags %s with %s", existing.inspect, new.inspect) ###
		bits, mask = new
		result = (existing & mask) | (bits & ~mask)
		#printf(" --> %s\n", result.inspect) ###
		return result
	end
	
	def Marathon.get_floor_ceiling_attributes(face, create = false)
		return get_attributes(face, FloorCeilingAttrs, create)
	end

	def Marathon.get_polygon_attributes(floor_face, create = false)
		return get_attributes(floor_face, PolygonAttrs, create)
	end

	def Marathon.get_platform_attributes(floor_face, create = false)
		return get_attributes(floor_face, PlatformAttrs, create)
	end
	
	def Marathon.get_wall_attributes(wall_face, create = false)
		return get_attributes(wall_face, SideAttrs, create)
	end
	
	def Marathon.get_object_attributes(obj, create = false)
		return get_attributes(obj, ObjectAttrs, create)
	end
	
	def Marathon.get_light_attributes(light_obj, create = false)
		return get_attributes(light_obj, LightAttrs, create)
	end

	def Marathon.get_media_attributes(light_obj, create = false)
		return get_attributes(light_obj, MediaAttrs, create)
	end

	def Marathon.get_ambient_sound_attributes(obj, create = false)
		return get_attributes(obj, AmbientSoundAttrs, create)
	end

	def Marathon.get_random_sound_attributes(obj, create = false)
		return get_attributes(obj, RandomSoundAttrs, create)
	end
	
	def Marathon.get_note_attributes(obj, create = false)
		return get_attributes(obj, NoteAttrs, create)
	end
	
	def Marathon.get_level_attributes(create = false)
		model = Sketchup.active_model()
		return get_attributes(model, LevelAttrs, create)
	end
	
	def Marathon.get_monster_placement_attributes(create = false)
		model = Sketchup.active_model()
		return get_attributes(model, MonsterPlacementAttrs, create)
	end

	def Marathon.get_item_placement_attributes(create = false)
		model = Sketchup.active_model()
		return get_attributes(model, ItemPlacementAttrs, create)
	end
	
	def Marathon.get_parameters(obj, create = false)
		return get_attributes(obj, ParameterAttrs, create)
	end

	#----------------------------------------------------------------------------------
	#
	#   Other attribute-related utilities
	#
	#----------------------------------------------------------------------------------

	def Marathon.get_control_panel_types()
		return ControlPanelDefinitions
	end

end # module Marathon
